// scatterplot.c - function to plot data using scatterplots for XmdvTool
//
// Generate scatterplots by dividing drawing area into N by N grid and draw
// each data sample N * N times.  Draw grids, vertical and horizontal labels,
// and grids.  Redraw bits under brush if state set to ON

#include <xmdv/CFlatScatter.h>
#include <xmdv/ogl.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/brush.h>
#include <xmdv/brush_oper.h>
#include <xmdv/multi_select.h>
#include <string.h>

CFlatScatter::CFlatScatter(CFlatDisplayInformation* flat_info) : CFlatDisplay(flat_info)
{
}


//
// Redraw a single point on the scatterplot
// display.  Useful for multi-sel.
//
//jing this func is used by canvas->.. _proc
void CFlatScatter::RedrawSingleScatter(grCanvas *canvas, int point)
{
//jing get data
    ARR<double> data(dataset->dims);
    dataset->GetData(data, point);

//jing set point size
//    glPointSize((float) ((DSIZE+1)*(canvas->zoom.X+canvas->zoom.Y)/2));
    glPointSize((float) ((DSIZE+1)));

//jing draw a point in scatterplot matrix ( in dim*dim plots )
    canvas->BeginDraw();
    glBegin(GL_POINTS);

    for (int j = 0;j < dataset->dims;j++)
        for (int k = 0;k < dataset->dims;k++)
            glVertex2f((float) CFlatScatter::XMAP(j,data[j]), (float) CFlatScatter::YMAP(k,data[k]) );

    glEnd();
//jing reset point size
    glPointSize(1.0);
}

//
// Draw an average value on the scatterplot
// display.
//
//jing I will not use this func temporarily
void CFlatScatter::DrawScatterAverage(grCanvas *canvas, const ARR<double> &data)
{
//    glPointSize((float) (3*(DSIZE+1)*(canvas->zoom.X+canvas->zoom.Y)/2));
    glPointSize((float) (3*(DSIZE+1)));
    canvas->BeginDraw();
    glBegin(GL_POINTS);
    for (int j=0; j<dataset->dims; j++)
        for (int k=0; k<dataset->dims; k++)
            glVertex2f((float) CFlatScatter::XMAP(j,data[j]), (float) CFlatScatter::YMAP(k,data[k]) );
    glEnd();
//jing why not in the last func? 
    canvas->EndDraw();
    glPointSize(1.0);
}


//
// Draw the brush boundary in the scatterplot
// view.
//
//jing this func is used by a func used by canvas->.. _proc
void CFlatScatter::DrawScatterBrushBoundRampOutline(grCanvas *canvas, Brush *brush)
{
    canvas->SetXorMode(true);

    for (int j=0; j<dataset->dims; j++) {
        for (int k=0; k<dataset->dims; k++)
        {
            Vec2 p, q;

            p.X = brush->RampMin(j);
            p.Y = brush->RampMax(k);

            p.X = CLAMP(p.X, dataset->dim_min[j], dataset->dim_max[j]);
            p.Y = CLAMP(p.Y, dataset->dim_min[k], dataset->dim_max[k]);

            q.X = brush->RampMax(j);
            q.Y = brush->RampMin(k);

            q.X = CLAMP(q.X, dataset->dim_min[j], dataset->dim_max[j]);
            q.Y = CLAMP(q.Y, dataset->dim_min[k], dataset->dim_max[k]);

            p.X = CFlatScatter::XMAP(j, p.X);
            p.Y = CFlatScatter::YMAP(k, p.Y);
            q.X = CFlatScatter::XMAP(j, q.X);
            q.Y = CFlatScatter::YMAP(k, q.Y);

            canvas->DrawRectangle(p.X, p.Y, q.X-p.X, q.Y-p.Y);
        }
    }
    canvas->SetXorMode(false);
}	

//draw the ramp around the brush
void CFlatScatter::DrawScatterBrushBoundRamp(grCanvas *canvas, Brush *brush, unsigned long color1, unsigned long color2)
{
    const double factor = 0.1;
    RGBt c1(color1), c2(color2);
    c2.R = factor*c1.R + (1-factor)*c2.R;
    c2.G = factor*c1.G + (1-factor)*c2.G;
    c2.B = factor*c1.B + (1-factor)*c2.B;

    canvas->BeginDraw();

    for (int j=0; j<dataset->dims; j++) {
        for (int k=0; k<dataset->dims; k++)
        {
            Vec2 p0,p1,q0,q1;

            p0.X = brush->Min(j);
            p0.Y = brush->Max(k);
            p1.X = p0.X - brush->bound_offset[j];
            p1.Y = p0.Y + brush->bound_offset[k];

            p1.X = CLAMP(p1.X, dataset->dim_min[j], dataset->dim_max[j]);
            p1.Y = CLAMP(p1.Y, dataset->dim_min[k], dataset->dim_max[k]);

            q0.X = brush->Max(j);
            q0.Y = brush->Min(k);
            q1.X = q0.X + brush->bound_offset[j];
            q1.Y = q0.Y - brush->bound_offset[k];

            q1.X = CLAMP(q1.X, dataset->dim_min[j], dataset->dim_max[j]);
            q1.Y = CLAMP(q1.Y, dataset->dim_min[k], dataset->dim_max[k]);

            p0.X = XMAP(j, p0.X);
            p0.Y = YMAP(k, p0.Y);
            p1.X = XMAP(j, p1.X);
            p1.Y = YMAP(k, p1.Y);
            q0.X = XMAP(j, q0.X);
            q0.Y = YMAP(k, q0.Y);
            q1.X = XMAP(j, q1.X);
            q1.Y = YMAP(k, q1.Y);

            glBegin(GL_QUAD_STRIP);

            glColor3f((float) c1.R, (float) c1.G, (float) c1.B);
            glVertex2f((float) p0.X, (float) p0.Y);
            glColor3f((float) c2.R, (float) c2.G, (float) c2.B);
            glVertex2f((float) p1.X, (float) p1.Y);
            glColor3f((float) c1.R, (float) c1.G, (float) c1.B);
            glVertex2f((float) p0.X, (float) q0.Y);
            glColor3f((float) c2.R, (float) c2.G, (float) c2.B);
            glVertex2f((float) p1.X, (float) q1.Y);
            glColor3f((float) c1.R, (float) c1.G, (float) c1.B);
            glVertex2f((float) q0.X, (float) q0.Y);
            glColor3f((float) c2.R, (float) c2.G, (float) c2.B);
            glVertex2f((float) q1.X, (float) q1.Y);
            glColor3f((float) c1.R, (float) c1.G, (float) c1.B);
            glVertex2f((float) q0.X, (float) p0.Y);
            glColor3f((float) c2.R, (float) c2.G, (float) c2.B);
            glVertex2f((float) q1.X, (float) p1.Y);
            glColor3f((float) c1.R, (float) c1.G, (float) c1.B);
            glVertex2f((float) p0.X, (float) p0.Y);
            glColor3f((float) c2.R, (float) c2.G, (float) c2.B);
            glVertex2f((float) p1.X, (float) p1.Y);

            glEnd();
        }
    }
    canvas->EndDraw();
}

//
// Given x and y dimensions, coordinates of the
// brush center, and brush size in each dimesnion
// draw the corresponding brush outline in xor mode
// on the scatterplot display.
//
//jing this func is used by funcs used by canvas->.. _proc
void CFlatScatter::DrawScatterBrushBox(grCanvas *canvas, int dimx, int dimy, double data_x, double data_y,
             double size_x, double size_y)
{
    Vec2 p, q;

    // lower left
    p.X = CFlatScatter::XMAP(dimx, data_x - size_x/2);
    p.Y = CFlatScatter::YMAP(dimy, data_y - size_y/2);

    // upper right
    q.X = CFlatScatter::XMAP(dimx, data_x + size_x/2);
    q.Y = CFlatScatter::YMAP(dimy, data_y + size_y/2);

    canvas->SetXorMode(true);
    canvas->DrawRectangle(p.X,p.Y,q.X-p.X,q.Y-p.Y);
    canvas->SetXorMode(false);
}

// Draw a single brush.
void CFlatScatter::DrawSingleBrush(Brush *brush)
{
    canvas->SetForeground( interface_color[BRUSH1+brush->num]);
    canvas->SetXorMode(false);
    for (int j=0; j<dataset->dims; j++) {
        for(int k=0; k<dataset->dims; k++)
        {
            Vec2 p, q;

            p.X = brush->Min(j);
            p.Y = brush->Max(k);

            p.X = CLAMP(p.X, dataset->dim_min[j], dataset->dim_max[j]);
            p.Y = CLAMP(p.Y, dataset->dim_min[k], dataset->dim_max[k]);

            q.X = brush->Max(j);
            q.Y = brush->Min(k);

            q.X = CLAMP(q.X, dataset->dim_min[j], dataset->dim_max[j]);
            q.Y = CLAMP(q.Y, dataset->dim_min[k], dataset->dim_max[k]);

            p.X = CFlatScatter::XMAP(j, p.X);
            p.Y = CFlatScatter::YMAP(k, p.Y);
            q.X = CFlatScatter::XMAP(j, q.X);
            q.Y = CFlatScatter::YMAP(k, q.Y);

            canvas->FillRectangle(p.X, p.Y, q.X-p.X, q.Y-p.Y);
        }
    }

    // draw boundary coverage if necessary
    if (brush->bound_type != Brush::STEP)
        DrawScatterBrushBoundRamp(canvas,brush,interface_color[BRUSH1+brush->num],interface_color[BG]);
}


//draw grid
void CFlatScatter::DrawAxes()
{
    if ( show_two_axes )
        DrawScatterGridWithTwoAxes(canvas);
    else if ( show_wide_axes )
        DrawScatterGridWideAxes(canvas);
    else
        DrawScatterGrid(canvas);
}


//set color and style for grid
void CFlatScatter::SetUnbrushedColorAndStyle()
{
    CFlatDisplay:: SetUnbrushedColorAndStyle();
//    glPointSize((float) ((DSIZE+1)*(canvas->zoom.X+canvas->zoom.Y)/2));
    glPointSize((float) ((DSIZE+1)));
    canvas->SetXorMode(false);
    glShadeModel(GL_FLAT);  // Because each line is the same color.
}


//draw single data
void CFlatScatter::DrawSingleData(ARR<double> data, int datano)
{
 //   canvas->BeginDraw();
    glBegin(GL_POINTS);

    for (int j=0; j<dataset->dims; j++) 
        for (int k=0; k<dataset->dims; k++)
            glVertex2f( (float) XMAP(j,data[j]), (float) YMAP(k,data[k]) );

    glEnd();
//    canvas->EndDraw();
}

void CFlatScatter::DrawSingleDataWithTwoAxes(ARR<double> mean, ARR<double> max, 
                                             ARR<double> min, int datano, bool unbrushed)
{
    glBegin(GL_POINTS);

    int j, k;

    for (j=0; j<dataset->dims; j++) 
    {
        for (k=0; k<dataset->dims; k++)
        {
            if ( j != k )
                glVertex2f((float) XMAP(j,mean[j]), (float) YMAP(k,mean[k]) );
        }
    }
    for (j=0; j<dataset->dims; j++)
    {
        glVertex2f((float) XMAP(j,min[j]), (float) YMAP(j,max[j]) );
    }
    
    
    glEnd();
}


//jing this func is used by canvas->.._proc
void CFlatScatter::InputScatterResizeBrush(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // Map mouse x/y location to data coordinates
    double mx = CFlatScatter::XINVMAP(dimx, p.X);
    double my = CFlatScatter::YINVMAP(dimy, p.Y);

    // If this is a diagonal plot (same dimension on both axes)
    // then average the x and y positions
    //
    if (dimx == dimy) mx = my = .5*(mx+my);

    // Bound mouse to plot edges
    //
    mx = MIN2(mx, dataset->dim_max[dimx]);
    mx = MAX2(mx, dataset->dim_min[dimx]);
    my = MIN2(my, dataset->dim_max[dimy]);
    my = MAX2(my, dataset->dim_min[dimy]);

    // Check to make sure min<max
    if (resize_minx && mx >= curr_brush->Max(dimx))
        mx = curr_brush->Max(dimx);
    if (resize_maxx && mx <= curr_brush->Min(dimx))
        mx = curr_brush->Min(dimx);
    if (resize_miny && my >= curr_brush->Max(dimy))
        my = curr_brush->Max(dimy);
    if (resize_maxy && my <= curr_brush->Min(dimy))
        my = curr_brush->Min(dimy);

    // Find the new bounds based on what is being resized
    //
    double new_minx, new_maxx, new_miny, new_maxy;

    if (resize_minx) new_minx = mx;
    else new_minx = curr_brush->Min(dimx);

    if (resize_maxx) new_maxx = mx;
    else new_maxx = curr_brush->Max(dimx);

    if (resize_miny) new_miny = my;
    else new_miny = curr_brush->Min(dimy);

    if (resize_maxy) new_maxy = my;
    else new_maxy = curr_brush->Max(dimy);

    if (resize_minx || resize_maxx)
    {
        curr_brush->SetMinMax(dimx,new_minx,new_maxx);
    }
    if (resize_miny || resize_maxy)
    {
        curr_brush->SetMinMax(dimy,new_miny,new_maxy);
    }

    canvas->SetXorMode(true);

    // Do the graph being manipulated
    DrawScatterBrushBox(canvas, dimx, dimy,
                  (new_maxx + new_minx)/2,
                  (new_maxy + new_miny)/2,
                  (new_maxx - new_minx),
                  (new_maxy - new_miny));

    // Do all other graphs along the horizontal
    for (int x=0; x<dataset->dims; x++)
    {
        if (x==dimx) continue;

        DrawScatterBrushBox(canvas, x, dimy,
                  curr_brush->pos[x],
                  (new_maxy + new_miny)/2,
                  curr_brush->size[x],
                  (new_maxy - new_miny));
    }

    // Do all other graphs along the vertical
    for (int y=0; y<dataset->dims; y++)
    {
        if (y==dimy) continue;

        DrawScatterBrushBox(canvas, dimx, y,
                  (new_maxx + new_minx)/2,
                  curr_brush->pos[y],
                  (new_maxx - new_minx),
                  curr_brush->size[y]);
    }

    canvas->SetXorMode(false);
    xor_needs_undo = true;
}


//jing this func is used by canvas->.._proc
void CFlatScatter::InputScatterResizeBoundRamp(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // Map mouse x/y location to data coordinates
    double mx = CFlatScatter::XINVMAP(dimx, p.X);
    double my = CFlatScatter::YINVMAP(dimy, p.Y);

    // If this is a diagonal plot (same dimension on both axes)
    // then average the x and y positions
    //
    if (dimx == dimy) mx = my = .5*(mx+my);

    // Bound mouse to plot edges
    //
    mx = MIN2(mx, dataset->dim_max[dimx]);
    mx = MAX2(mx, dataset->dim_min[dimx]);
    my = MIN2(my, dataset->dim_max[dimy]);
    my = MAX2(my, dataset->dim_min[dimy]);

    // Check to make sure min<max
    if (resize_minx && mx >= curr_brush->Max(dimx))
        mx = curr_brush->Max(dimx);
    if (resize_maxx && mx <= curr_brush->Min(dimx))
        mx = curr_brush->Min(dimx);
    if (resize_miny && my >= curr_brush->Max(dimy))
        my = curr_brush->Max(dimy);
    if (resize_maxy && my <= curr_brush->Min(dimy))
        my = curr_brush->Min(dimy);

    if (resize_minx)
    {
        if (mx > curr_brush->Min(dimx))
            curr_brush->bound_offset[dimx] = 0.0;
        else
            curr_brush->bound_offset[dimx] = curr_brush->Min(dimx) - mx;
    }

    if (resize_maxx)
    {
        if (mx < curr_brush->Max(dimx))
            curr_brush->bound_offset[dimx] = 0.0;
        else
            curr_brush->bound_offset[dimx] = mx - curr_brush->Max(dimx);
    }

    if (resize_miny)
    {
        if (my > curr_brush->Min(dimy))
            curr_brush->bound_offset[dimy] = 0.0;
        else
            curr_brush->bound_offset[dimy] = curr_brush->Min(dimy) - my;
    }

    if (resize_maxy)
    {
        if (my < curr_brush->Max(dimy))
            curr_brush->bound_offset[dimy] = 0.0;
        else
            curr_brush->bound_offset[dimy] = my - curr_brush->Max(dimy);
    }

    DrawScatterBrushBoundRampOutline(canvas,curr_brush);

    xor_needs_undo = true;
}

//jing this func is used by canvas->.. _proc
void CFlatScatter::InputScatterMoveBrush(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // Map mouse x/y location to data coordinates
    double mx = CFlatScatter::XINVMAP(dimx, p.X);
    double my = CFlatScatter::YINVMAP(dimy, p.Y);

    // If this is a diagonal plot (same dimension on both axes)
    // then average the x and y positions
    //
    if (dimx == dimy) mx = my = .5*(mx+my);

    // Bound mouse to plot edges
    //
    mx = MIN2(mx, dataset->dim_max[dimx] - curr_brush->size[dimx]/2);
    mx = MAX2(mx, dataset->dim_min[dimx] + curr_brush->size[dimx]/2);

    my = MIN2(my, dataset->dim_max[dimy] - curr_brush->size[dimy]/2);
    my = MAX2(my, dataset->dim_min[dimy] + curr_brush->size[dimy]/2);

    curr_brush->pos[dimx] = mx;
    curr_brush->pos[dimy] = my;

    canvas->SetXorMode(true);

    DrawScatterBrushBox(canvas, dimx, dimy, mx, my, curr_brush->size[dimx], curr_brush->size[dimy]);

    // Do all other graphs along the horizontal
    for (int x=0; x<dataset->dims; x++)
    {
        if (x==dimx) continue;

        DrawScatterBrushBox(canvas, x, dimy, curr_brush->pos[x], my,
            curr_brush->size[x], curr_brush->size[dimy]);
    }

    // Do all other graphs along the vertical
    for (int y=0; y<dataset->dims; y++)
    {
        if (y==dimy) continue;

        DrawScatterBrushBox(canvas, dimx, y, mx, curr_brush->pos[y],
            curr_brush->size[dimx], curr_brush->size[y]);
    }

    canvas->SetXorMode(false);
    xor_needs_undo = true;
}

//jing this func is used by canvas->.. _proc
void CFlatScatter::InputScatterMultiSelect(grCanvas *canvas, int sx, int sy)
{
    if (!multi_select) return;

    ARR<double> data(dataset->dims);

    // Find all points within circle distance of mouse
    //
    for (int i=0; i<dataset->data_size; i++)
    {
        dataset->GetData(data, i);

        Vec2 p;
        p.X = CFlatScatter::XMAP(dimx,data[dimx]);
        p.Y = CFlatScatter::YMAP(dimy,data[dimy]);
        Vec2 q = canvas->MapDataToScreen(p);

        const int rad2 = multisel_radius*multisel_radius; // radius^2 in screen coords

        if ((q.X-sx)*(q.X-sx) + (q.Y-sy)*(q.Y-sy) < rad2)
        {
            multiSelectAddPoint(i);
            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[PAINT]);
            RedrawSingleScatter(canvas, i);
            glDrawBuffer(GL_BACK);
        }
    }
}


//jing this func is used by canvas->.. _proc
void CFlatScatter::FindResizeCorners(double x,double y)
{
    // There are four different variables that could be resized
    // (minx, maxx, miny, maxy) the user can resize at most
    // two at once

    // Reset all resize flags
    resize_minx = resize_maxx = false;
    resize_miny = resize_maxy = false;

    // Map mouse x/y location to data coordinates
    double mx = CFlatScatter::XINVMAP(dimx, x);
    double my = CFlatScatter::YINVMAP(dimy, y);

    // If this is a diagonal plot (same dimension on both axes)
    // then average the x and y positions
    //
    if (dimx == dimy) mx = my = .5*(mx+my);

    // Check if user is resizing in x direction
    if (mx <= curr_brush->pos[dimx] - .25*curr_brush->size[dimx])
        resize_minx = true;
    else
    if (mx >= curr_brush->pos[dimx] + .25*curr_brush->size[dimx])
        resize_maxx = true;

    // Check if user is resizing in y direction
    if (my <= curr_brush->pos[dimy] - .25*curr_brush->size[dimy])
        resize_miny = true;
    else
    if (my >= curr_brush->pos[dimy] + .25*curr_brush->size[dimy])
        resize_maxy = true;
}


void CFlatScatter::InputMouseMotion(int x, int y)
{
    if (!dataset) return;

    if ( screen_distortion->distortion_status )
    {
        InputScatterFindPlot(canvas, x, y, dimx, dimy);
        return;
    }

    ScatterCaption(canvas, x, y);
    static int prev_sx, prev_sy;

    switch (input_mode) {
        case MULTI_SELECT:
            InputScatterMultiSelect(canvas,x,y);
            return;

        case RESIZE_BRUSH:
            if (xor_needs_undo) InputScatterResizeBrush(canvas,prev_sx,prev_sy);
            InputScatterResizeBrush(canvas,x,y);
            break;

        case MOVE_BRUSH:
            if (xor_needs_undo) InputScatterMoveBrush(canvas,prev_sx,prev_sy);
            InputScatterMoveBrush(canvas,x,y);
            break;

        case RESIZE_RAMP:
            if (xor_needs_undo) InputScatterResizeBoundRamp(canvas,prev_sx,prev_sy);
            InputScatterResizeBoundRamp(canvas,x,y);
            break;

        case CANVAS_PAN:
            appCanvasMousePan(canvas);
            break;

        default: break;
    }
    prev_sx = x;
    prev_sy = y;
}


void CFlatScatter::InputButtonPress(int button, int x, int y)
{
    if (!dataset) return;

    if ( screen_distortion->distortion_status )
    {
        return;
    }

    if (canvas->button[2]==true) {
        input_mode = CANVAS_PAN;
        canvas->SetHandCursor();
        return;
    }

    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );
/*
    double fx = p.X * dataset->dims;
    double fy = p.Y * dataset->dims;
    if (fx<0 || fy<0) return;

    dimx = int(fx);
    dimy = int(fy);
*/
    InputScatterFindPlot(canvas, x, y, dimx, dimy);

    if (dimx < 0 || dimx >= dataset->dims || dimy < 0 || dimy >= dataset->dims) {
        dimx = dimy = -1;
        return;
    }

    if (canvas->shiftkey) {
        input_mode = MULTI_SELECT;
        InputScatterMultiSelect(canvas,x,y);
        return;
    } else
    if (canvas->controlkey && curr_brush->bound_type == Brush::RAMP && button == 0) {
        input_mode = RESIZE_RAMP;
        FindResizeCorners(p.X,p.Y);
        InputScatterResizeBoundRamp(canvas,x,y);
        return;
    }

    assert(xor_needs_undo==false);

    if (canvas->button[0] == true) {
        input_mode = RESIZE_BRUSH;
        FindResizeCorners(p.X,p.Y);
    } else
    if (canvas->button[1] == true) {
        input_mode = MOVE_BRUSH;
    }

    InputMouseMotion(x,y);
}

void CFlatScatter::InputButtonRelease(int button, int x, int y)
{
    if (!dataset) return;

    if ( screen_distortion->distortion_status )
    {
        double delta;

        input_mode = INCREASE_PLOT_SIZE;
        switch (input_mode) 
        {
            case INCREASE_PLOT_SIZE:
                delta = 1. / dataset->dims * 0.2;
                InputScatterScalePlotSize(canvas, dimx, dimy, delta);
                canvas->Refresh();
                break;
            case DECREASE_PLOT_SIZE:
                delta = (-1.) / dataset->dims * 0.2;
                InputScatterScalePlotSize(canvas, dimx, dimy, delta);
                break;
            default: break;
        }
        return;
    }

    switch (input_mode) {
        case MULTI_SELECT:
            break;

        case CANVAS_PAN:
            canvas->SetDefaultCursor();
            break;

        case RESIZE_RAMP:
        case RESIZE_BRUSH:
        case MOVE_BRUSH:
            canvas->SetDefaultCursor();
            BrushOperMarkDataset();
            appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
            break;

        default: break;
    }

    input_mode = NONE;
    xor_needs_undo = false;
}


void CFlatScatter::SetAverageColorAndStyle()
{
    CFlatDisplay::SetAverageColorAndStyle();
//    glPointSize((float) (3*(DSIZE+1)*(canvas->zoom.X+canvas->zoom.Y)/2));
    glPointSize((float) (3*(DSIZE+1)));

}


bool CFlatScatter::TestDimensionality()  
{
	return ScatterTestDimensionality();
}