
#include <xmdv/brush_oper.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/cluster.h>
#include <string.h>
#include <stdio.h>

#ifndef WIN32
#include <unistd.h> // for Linux sleep function
#endif

BrushOperation brush_ops[MAXOPER]; // the operations
int num_oper=0;         // number of operations
int list_edit;          // the list to edit

char *oper_names[MAXOPER] = {NULL}; // the operation names

//
// Initialize the brush operations list
//
void BrushOperListInit()
{
    // Initialize one highlight operation for the first brush
    num_oper = 1;
    brush_ops[0].num_brush = 1;
    brush_ops[0].xnot[0] = 0;
    brush_ops[0].brush[0] = 0;
    brush_ops[0].oper = HIGH;
    BITSET(brush_ops[0].oper,VALUES); 
/*
    num_oper = 2;
    brush_ops[0].num_brush = 1;
    brush_ops[0].xnot[0] = 0;
    brush_ops[0].brush[0] = 0;
    brush_ops[0].oper = HIGH;

    brush_ops[1].num_brush = 1;
    brush_ops[1].xnot[0] = 1;
    brush_ops[1].brush[0] = 0;
    brush_ops[1].oper = MASK;
*/
}

//
// Instantiate a new brush. Returns the new operation number.
//
int BrushOperNew()
{
    num_oper++;
    assert(num_oper<MAXOPER /* Can only support max of 30 operations */ );
    return num_oper-1;
}

//
// Delete a brush
//
void BrushOperDelete(int which)
{
    for (int i=which+1; i<num_oper; i++)
        brush_ops[i-1] = brush_ops[i];

    num_oper--;
    if (num_oper<0) num_oper=0;
}

//
// Lower priority of brush
//
void BrushOperLowerPriority(int which)
{
    if (which == num_oper-1) return;

    // change the priority
    BrushOperation tmp = brush_ops[which];
    brush_ops[which] = brush_ops[which+1];
    brush_ops[which+1] = tmp;
}

//
// Raise priority of brush
//
void BrushOperRaisePriority(int which)
{
    if (which == 0) return;

    BrushOperation tmp = brush_ops[which];
    brush_ops[which] = brush_ops[which-1];
    brush_ops[which-1] = tmp;
}

//
// Generate an ASCII name for an operation from
// an Operation structure.
//
const char *BrushOperGenerateString(int which)
{
    static char buf[512];

    BrushOperation *op = &brush_ops[which];

    // Initialize buffer
    sprintf(buf, "%2d: ", which+1);

    for (int i=0; i<op->num_brush; i++)
    {
        // Add a NOT if necessary
        if (op->xnot[i]) strcat(buf, "NOT ");

        // Add the brush name
        char buf2[256];
        sprintf(buf2, "B%d ", op->brush[i]+1);
        strcat(buf, buf2);

        // Add a conjunction if there are more brushes
        if (i < op->num_brush-1) {
            switch(op->conj[i])
            {
            case AND:
                strcat(buf, "AND ");
                break;
            case OR:
                strcat(buf, "OR ");
                break;
            case XOR:
                strcat(buf, "XOR ");
                break;
            }
        }
    }
    if (op->num_brush==0) strcat(buf,"<no brush> ");

    strcat(buf,": ");

    // Now add the operation(s)
    if (op->oper & HIGH)
        strcat(buf, "highlight ");

    if (op->oper & MASK)
        strcat(buf, "mask ");

    if (op->oper & VALUES)
        strcat(buf, "values ");

    if (op->oper & AVERAGE)
        strcat(buf, "average ");

    if (op->oper == 0) {
        strcat(buf,"<null operation>");
    }

    // copy the name and return it
    return buf;
}

//
// Compute if a point is covered by the given brush
// operation.
//
bool BrushOperCheckPoint(const ARR<double> &point, int which)
{
    BrushOperation *op = &brush_ops[which];

    // Check to make sure all brushes are enabled first
    int i;
    for (i=0; i<op->num_brush; i++) {
        int brush_num = op->brush[i];
        if (brushes[brush_num].enable == false)
            return false;
    }
    // Set the result to the initial brush coverage
    bool result = op->xnot[0] ? !CheckBrushCoverage(point, &brushes[op->brush[0]]) :
                           CheckBrushCoverage(point, &brushes[op->brush[0]]);

    for (i=1; i<op->num_brush; i++)
    {
        bool next_br = op->xnot[i] ? !CheckBrushCoverage(point, &brushes[op->brush[i]]) :
                                CheckBrushCoverage(point, &brushes[op->brush[i]]);

        switch(op->conj[i-1])
        {
        case AND:
            result &= next_br;
            break;
        case OR:
            result |= next_br;
            break;
        case XOR:
            result ^= next_br;
            break;
        }
    }
    return result;
}

//
// Compute the percent of coverage from a brush.  Only really makes sense for
// brushes with non stepped boundaries. Returns the coverage percent (0..1).
//
static double PercentOperationPoint(const ARR<double> &point, int which)
{
    BrushOperation *op = &brush_ops[which];

    // Check to make sure all brushes are enabled first
    int i;
    for (i=0; i<op->num_brush; i++)
        if (brushes[op->brush[i]].enable == false)
            return 0.;

    // Set the result to the initial brush coverage
    double coverage = PercentBrushCoverage(point, &brushes[op->brush[0]]);
    if (op->xnot[0]) {
        assert( 0 /* Not till multiops is ported */ );
        coverage = 1-coverage;
    }

    for (i=1; i<op->num_brush; i++)
    {
        assert( 0 /* Not till multiops is ported */ );
        double next_br = PercentBrushCoverage(point, &brushes[op->brush[i]]);
        if (op->xnot[i]) next_br = 1-next_br;

        switch(op->conj[i-1])
        {
        case AND:
            coverage = MIN2(coverage, next_br);
            break;
        case OR:
            coverage = MAX2(coverage, next_br);
            break;
        case XOR:
            coverage += next_br;
            if (coverage > 1.0) coverage = 2.0 - coverage;
            break;
        }
    }

    return coverage;
}

//
// Quick check to see if point is out of min/max range
//
static bool ExceedsDimensionBoundaries(const ARR<double> &data)
{
    for (int i=0;i<dataset->dims;i++) {
        if (data[i]<dataset->dim_min[i]) return true;
        if (data[i]>dataset->dim_max[i]) return true;
    }
    return false;
}

//
// This function determines the color that a point should be from the various
// brush_ops defined. The lowest numbered operation has precedence if the
// point is covered by multiple brush_ops. Returns the color cell to paint the
// point with.
//
static void BrushOperComputeColor(const ARR<double> &data, okcdata::mark_type &what_type, RGBt &what_color)
{
    // If the point is out of the dim_min/max range,
    // disable it (this happens during zooming)
    //
    if (ExceedsDimensionBoundaries(data)) {
        what_type = okcdata::DISABLED;
        return;
    }

    // If point is not covered by any operations, just
    // paint it in the data color.
    //
    what_type = okcdata::REGULAR;
    what_color = interface_color[DATA];

    // The first operation has highest priority
    //
    for (int i=num_oper-1; i>=0; i--)
    {
        if (brush_ops[i].num_brush == 0) continue;

        bool enclosed = BrushOperCheckPoint(data, i);
        if (!enclosed) continue;

        if (brush_ops[i].oper & HIGH)
        {
            what_type = okcdata::COLORED;

// NEED TO RETHINK THIS CONDITION....
            if (brush_ops[i].num_brush == 1 && brushes[brush_ops[i].brush[0]].bound_type==Brush::RAMP) 
            {
                // Can't resolve multiple ramp brushes ?!?
                double coverage = PercentOperationPoint(data,i);
                assert(coverage > 0);
                RGBt cdata = interface_color[BG]; // [DATA]?
                RGBt chigh = interface_color[HIGHLIGHT1+i];

                if (coverage < 1.0) coverage -= 0.5; // HACK : To make fading more obvious

                coverage = CLAMP(coverage,0.0,1.0);
                RGBt color;
                color.R = (coverage)*chigh.R + (1-coverage)*cdata.R;
                color.G = (coverage)*chigh.G + (1-coverage)*cdata.G;
                color.B = (coverage)*chigh.B + (1-coverage)*cdata.B;
                what_color = color;

//                return;
            }

            what_color = interface_color[HIGHLIGHT1+i];
//            return;
        } else
        if (brush_ops[i].oper & MASK) {
            what_type = okcdata::DISABLED;
//            return;
        }
    }

}

//start_jing_animate_brush

static double fadeout(double old_time)
{
    if ( old_time < 0.01 )
        old_time = 0;
    else
        old_time /= 1.1;
    return old_time;
}

static void AnimatedBrushOperComputeColor(double *time, const ARR<double> &data, okcdata::mark_type &what_type, RGBt &what_color)
{
    // If the point is out of the dim_min/max range,
    // disable it (this happens during zooming)
    //
    if (ExceedsDimensionBoundaries(data)) {
        what_type = okcdata::DISABLED;
        return;
    }

    if ( *time < 0.1 )
    {
        what_type = okcdata::REGULAR;
        what_color = interface_color[DATA];
    }

    bool enclosed = CheckBrushCoverage(data, curr_brush);
    if (!enclosed) 
    {
        //try1
      /*  *time = fadeout(*time);
        continue;
        */
        //try2
        if ( *time >= 0.1 )
        {
            *time = fadeout(*time);
            RGBt data_color = interface_color[DATA]; 
            what_color.R = (*time) * (what_color.R) + (1 - (*time)) * (data_color.R);
            what_color.G = (*time) * (what_color.G) + (1 - (*time)) * (data_color.G);
            what_color.B = (*time) * (what_color.B) + (1 - (*time)) * (data_color.B);
        }
    }
    else
    {
        what_type = okcdata::COLORED;
        *time = 1.;
        what_color = interface_color[HIGHLIGHT1];
    } 

}

//end_jing_animate_brush

//
// Scans all points in the dataset and assign each a color
// for drawing.
//
/*
void BrushOperMarkDataset()
{
    ARR<double> data(dataset->dims);

    int i;
    for (i=0;i<dataset->data_size;i++) {
        dataset->GetData(data,i);
        BrushOperComputeColor(data,dataset->mark[i],dataset->color[i]);
    }

    // Traverse the cluster tree and mark the clusters that are within the brush
    if (!btree) return;

    for (i=0;i<btree->nodes.Len();i++) {
        Cluster &c = btree->nodes[i];
        RGBt dummy_color;
        BrushOperComputeColor(c.mean,c.mark,dummy_color);
    }
}*/

//start_jing_animate_brush
void BrushOperMarkDataset()
{
    ARR<double> data(dataset->dims);

    if ( flat_info->flat_sb_brush)
    {
        if ( btree )
            hierMarkFlatBrushedClustersByExtents(curr_brush, &btree->nodes[0]);
    }
    else
    {

        int i;
        for (i=0;i<dataset->data_size;i++) {
            dataset->GetData(data,i);
            if ( dataset->animate_brush )
                AnimatedBrushOperComputeColor(&(dataset->brushed_time[i]), data,dataset->mark[i],dataset->color[i]);
            else
                BrushOperComputeColor(data,dataset->mark[i],dataset->color[i]);
        }
    }
}
//end_jing_animate_brush

//
// Given an operation number, computes the average
// of all the points contained by the operation and
// stores the result in aver.
//
int BrushOperCompAverage(int op, ARR<double> &aver)
{
    int i;
    for (i=0; i<dataset->dims; i++) aver[i] = 0.0;

    int count=0;
    ARR<double> data(dataset->dims);

    for (i=0; i<dataset->data_size; i++)
    {
        dataset->GetData(data,i);

        if (BrushOperCheckPoint(data, op))
        {
            for (int j=0; j<dataset->dims; j++) aver[j] += data[j];
            count++;
        }
    }

    if (count!=0) {
        for (i=0; i<dataset->dims; i++) aver[i] /= count;
    }

    return count;
}


//start_jing_animate_brush

#include <xmdv/ogl.h>
CLinkList *saved_brush_list = NULL;

static double InterpolateValue(double value1, double value2, int total, int now)
{
    double value;
    value = value1 + (value2 -  value1) / total * now;
    return value;
}

static void InterpolateBrush(Brush *brush1, Brush *brush2, Brush *brush, int total, int now)
{
    int dim = brush->pos.Len();

    for (int i = 0; i < dim; i++ )
    {
        brush->pos[i] = InterpolateValue(brush1->pos[i], brush2->pos[i], total, now);
        brush->size[i] = InterpolateValue(brush1->size[i], brush2->size[i], total, now);
        brush->bound_offset[i] = InterpolateValue(brush1->bound_offset[i], brush2->bound_offset[i], total, now);
    }
}

void AnimateBrush(grCanvas *canvas, Brush *brush1, Brush *brush2)
{
    if ( canvas->display != NULL )
    {
        Brush *brush = new Brush();
        brush->Copy(*brush1);
        int total = 10;
        int now = 1;
        for ( int i = 0; i < total; i++ )
        {
            curr_brush = brush; 
            BrushOperMarkDataset();
            canvas->Refresh(true);
#ifdef WIN32
            Sleep(200);
#else
	    sleep(200 / 1000);
#endif

            InterpolateBrush(brush1, brush2, brush, total, now);
            now ++;
        }
    }

}

void AnimateSavedBrushList(grCanvas *canvas)
{
    if ( saved_brush_list == NULL )
        return;

    dataset->animate_brush = true;
    Brush *old_curr_brush = curr_brush;
    Brush *brush1, *brush2;
    int number_saved_brushes = saved_brush_list->GetLength();
    if ( number_saved_brushes < 2 )
        return;

    brush1 = new Brush();
    brush2 = new Brush();
    CopyBrushFromAllBrush(brush1, (Brush *)(saved_brush_list->GetElement(0)));
    for ( int i = 1 ; i < number_saved_brushes; i++ )
    {
        CopyBrushFromAllBrush(brush2, (Brush *)(saved_brush_list->GetElement(i)));
        AnimateBrush(canvas, brush1, brush2);
        brush1->Copy(*brush2);
    }

    delete brush1;
    delete brush2;

    dataset->animate_brush = false;

    curr_brush = old_curr_brush;
    BrushOperMarkDataset();
}

void DeleteSavedBrushList()
{
    if ( saved_brush_list == NULL )
        return;

    int number_saved_brushes = saved_brush_list->GetLength();
    for ( int i = 0; i < number_saved_brushes; i++ )
    {
        delete (Brush *)(saved_brush_list->GetElement(i));
    }

    delete saved_brush_list;

    saved_brush_list = NULL;
}

void CopyBrushToAllBrush(Brush *all_brush, Brush *brush)
{
    all_brush->num = brush->num;
    all_brush->enable = brush->enable;
    all_brush->display = brush->display;
    all_brush->bound_type = brush->bound_type;

    int dim = brush->pos.Len();
    all_brush->pos.SetLen(dim);
    all_brush->size.SetLen(dim);
    all_brush->bound_offset.SetLen(dim);

    for ( int i = 0; i < dataset->all_dims; i++ )
    {
        all_brush->pos[i] = (dataset->all_dim_min[i] + dataset->all_dim_max[i]) / 2.;
        all_brush->size[i] = dataset->all_dim_max[i] - dataset->all_dim_min[i];
        all_brush->bound_offset[i] = (dataset->all_dim_max[i] - dataset->all_dim_min[i]) / 10.0;
    }
    dataset->ReverseMapData(brush->pos, all_brush->pos);
    dataset->ReverseMapData(brush->size, all_brush->size);
    dataset->ReverseMapData(brush->bound_offset, all_brush->bound_offset);
}

void CopyBrushFromAllBrush(Brush *brush, Brush *all_brush)
{
    brush->num = all_brush->num;
    brush->enable = all_brush->enable;
    brush->display = all_brush->display;
    brush->bound_type = all_brush->bound_type;

    int dim = all_brush->pos.Len();
    brush->pos.SetLen(dim);
    brush->size.SetLen(dim);
    brush->bound_offset.SetLen(dim);

    dataset->MapData(all_brush->pos, brush->pos);
    dataset->MapData(all_brush->size, brush->size);
    dataset->MapData(all_brush->bound_offset, brush->bound_offset);
}

void SaveCurrentBrushToSavedBrushList()
{
    if ( !dataset )
        return;

    if ( saved_brush_list == NULL )
        saved_brush_list = new CLinkList();

    Brush* all_brush = new Brush();
    CopyBrushToAllBrush(all_brush, curr_brush);
    saved_brush_list->AddNodeAtEnd(all_brush);
}

//end_jing_animate_brush
