// Dimension.cpp: implementation of the Dimension class.
//
//////////////////////////////////////////////////////////////////////

#pragma warning(disable:4786)

#include <iostream>
#include <math.h>	// for fabs
#include <assert.h>	// for assert

#include <SAX/InputSource.h>
#include <SAX/helpers/DefaultHandler.h>

#include <meta/Meta.h>
#include <meta/Dimension.h>

#include <meta/MetaHelper.h>

//////////////////////////////////////////////////////////////////////

struct IMapper;

class Dimension::DimensionHandler : public SAX::DefaultHandler
{
public:
	DimensionHandler(Dimension *dimension);

	SAX::InputSource input;
	ISerializable *parent;
	
	////////////////////////////////////////////////////
	// DefaultHandler
	virtual void startElement(const std::string& namespaceURI, 
		const std::string& localName,
		const std::string& qName, 
		const AttributesT& atts);
	virtual void endElement(const std::string& namespaceURI, 
		const std::string& localName,
		const std::string& qName);
	virtual void characters(const std::string& ch);

	virtual void fatalError(const SAX::SAXException& exception);
	
private:
	Dimension *mDimension;
	bool mHasValue;
	IMapper *mapper;

};

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

class Dimension::MappingIterator : public IMappingIterator
{
public:
	MappingIterator(Dimension *dimension) {
		mDimension = dimension;
		dimMapIter = dimension->mMappedValues.begin();
	}
	bool hasNext() const {
		return dimMapIter != mDimension->mMappedValues.end();
	}
	
	void next(float *fValue, IValue **vValue) {
		*fValue = dimMapIter->first;
		*vValue = dimMapIter->second;
		++dimMapIter;
	}

private:
	Dimension *mDimension;
	ValueMapCIter dimMapIter;
};

//////////////////////////////////////////////////////////////////////
// Mappers
//////////////////////////////////////////////////////////////////////

struct IMapper
{
	virtual float map() =0;
	virtual void  init(float f) =0;
	virtual ~IMapper() {}
};

class IncrementalMapper : public IMapper
{
public:
	IncrementalMapper() { init(0);			}
	float map()			{ return ++mapping;	}
	void init(float f)  { mapping = f;		}
private:
	float mapping;
};

class ScaledMapper : public IMapper
{
public:
	ScaledMapper(float f) { init(f);		}
	float map()			{ return scale;		}
	void init(float f)  { scale = f;		}
private:
	float scale;
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Dimension::Dimension()
{
	mHandler = new DimensionHandler(this);
}

Dimension::~Dimension()
{
	ValueMapIter i = mMappedValues.begin();
	for ( ; i != mMappedValues.end(); ++i) {
		IValue *value = i->second;
		delete value;
	}
	mMappedValues.clear();
}

void Dimension::setName(const char *name)
{
	mName = name;
}

const char* Dimension::getName() const
{
	return mName.c_str();
}

Type Dimension::getType() const
{
	return mType;
}

void Dimension::setType(Type t)
{
	mType = t;
}

IValue* Dimension::getMappedValue(float f) const
{
	ValueMapCIter i = mMappedValues.begin();
	for ( ; i != mMappedValues.end(); ++i) {
		if (fabs(i->first - f) < 0.5f) {
			return i->second;
		}
	}
	return 0;
}

IMappingIterator* Dimension::getMappingIterator()
{
	return new MappingIterator(this);
}

void Dimension::setMapping(float f,IValue *value)
{
	mMappedValues[f] = value;	
}

bool Dimension::equal()
{
	return false;
}

bool Dimension::unmarshal(std::istream &is, ISerializable *parent)
{
	METAHELPER->pushHandler(mHandler);

	if (parent == 0) {
		mHandler->input.setByteStream(is);
		
		if(mHandler->input.getByteStream() == 0)
			return false;
		METAHELPER->getParser().parse(mHandler->input);
	}
	
	return true;
}

bool Dimension::marshal(std::ostream& os, ISerializable *parent)
{
	os << "<dimension name=\"" << mName 
				<< "\" type=\"" << mType.getName() << "\"";
	if (mMappedValues.size() == 0) {
		os << " />" << std::endl;
	} else {
		os << ">" << std::endl;

		ValueMapCIter i = mMappedValues.begin();
		for ( ; i != mMappedValues.end(); ++i) {
			float f = i->first;
			IValue *value = i->second;

			os << "<value scale=\"" << f << "\">" << std::endl;
			os << value->getStringRepresentation() << std::endl;
			os << "</value>";
		}	

		os << "</dimension>" << std::endl;
	}
	
	return true;
}

//////////////////////////////////////////////////////////////////////////
// DimensionHandler

Dimension::DimensionHandler::DimensionHandler(Dimension *dimension)
: mDimension(dimension) 
{
	mHasValue = false; mapper = 0;
}

/*virtual*/ 
void Dimension::DimensionHandler::startElement(const std::string& namespaceURI,
											   const std::string& localName,
											   const std::string& qName,
											   const AttributesT& atts)
{
	if (localName.length()) {
		if (localName == "dimension") {
			
			std::string strName = atts.getValue("name");
			std::string strType = atts.getValue("type");
			
			mDimension->setName(strName.c_str());
			mDimension->setType(Type::fromName(strType.c_str()));
			
		} else if (localName == "value") {
//			std::cerr << "Start Elem: " << localName << std::endl;
			if (!mapper) {
				std::string dist = atts.getValue("scale");
				if (dist.length()) {
					mapper = new ScaledMapper(atof(dist.c_str()));
				} else {
					mapper = new IncrementalMapper();
				}
			} else {
				std::string dist = atts.getValue("scale");
				if (dist.length()) {
					mapper->init(atof(dist.c_str()));
				}	
			}
			mHasValue = true;
		}
	}
}

/*virtual*/ 
void Dimension::DimensionHandler::endElement(const std::string& namespaceURI,
										     const std::string& localName,
										     const std::string& qName)
{
	if (localName.length()) {
//		std::cerr << "End Elem: " << localName << std::endl;
		if (localName == "dimension") {
			delete mapper;
			mapper = 0;
			assert (this == METAHELPER->popHandler());
		}
	}
}

/*virtual*/
void Dimension::DimensionHandler::characters(const std::string& ch)
{
	if (mHasValue) {
		if (mDimension->mType.isBad())
			return;
		IValue *value;
		if (mDimension->mType.canCreateInstance()) {
			value = mDimension->mType.createInstance(ch.c_str());
			if (value) {
				mDimension->setMapping(mapper->map(), value);
			}
		}
		mHasValue = false;
	}
}

/*virtual*/ 
void Dimension::DimensionHandler::fatalError(const SAX::SAXException& exception)
{
	std::string e = exception.what();
}
